import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/nullability_suffix.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:analyzer/dart/element/visitor.dart';
import 'package:build/build.dart';
import 'package:controller/controller.dart';
import 'package:controller_generator/src/util.dart';
import 'package:source_gen/source_gen.dart';

const httpRequestMeta = {
  'Get',
  'Put',
  'Post',
  'Delete',
  'Options',
  'Patch',
  'Head',
  'HttpRequest',
};

typedef TypeCheck = bool Function(DartType);

const coreValidators = <TypeCheck, String>{
  isNum: 'number',
  isInt: 'integer',
  isDouble: 'decimal',
  isRequired: 'required',
};

bool isNum(DartType t) => t.isDartCoreNum;

bool isInt(DartType t) => t.isDartCoreInt;

bool isDouble(DartType t) => t.isDartCoreDouble;

bool isRequired(DartType t) => t.nullabilitySuffix == NullabilitySuffix.none;

class DispatcherBuilder extends GeneratorForAnnotation<Controller> {
  @override
  String generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    if (element.kind != ElementKind.CLASS) {
      throw 'ERROR: @Constructor can only be used on a class, found on $element';
    }
    var typeName = element.name;
    var matcherBuilder = MatcherBuilder();
    element.visitChildren(matcherBuilder);

    if (matcherBuilder.matchers.isEmpty) {
      return '';
    }
    return '''
    class $typeName\$DispatcherBuilder extends DispatcherBuilder {
      final $typeName controller;
      ${matcherBuilder.hasSecurity ? 'final Security security;' : ''}

      $typeName\$DispatcherBuilder(this.controller${matcherBuilder.hasSecurity ? ', this.security' : ''});
      
      @override
      AsyncPatternMatcher<Request, Response> appendMatchers(
          AsyncPatternMatcher<Request, Response> requestMatcher) {
        return requestMatcher
        ${matcherBuilder.matchers.join('\n')};
      }
      
      ${matcherBuilder.requestHandlers.join('\n\n')}
    }
    ''';
  }
}

class MatcherBuilder extends SimpleElementVisitor<void> {
  final List<String> matchers = [];
  final List<String> requestHandlers = [];
  bool hasSecurity = false;

  @override
  void visitMethodElement(MethodElement element) {
    var httpRequests = element.metadata.map(_extractHttpRequest).whereNotNull();
    var secured = _getMetadataOfType(
        element.metadata,
        'Secured',
        (meta, instance) =>
            SecurityInfo(instance.getField('statusCode')!.toIntValue()!, meta));
    if (secured != null) {
      hasSecurity = true;
    }

    matchers.addAll(
        httpRequests.map((httpRequest) => _buildMatcher(element, httpRequest)));
    requestHandlers.addAll(httpRequests.map(
        (httpRequest) => _buildRequestHandler(element, httpRequest, secured)));
  }

  T? _getMetadataOfType<T>(List<ElementAnnotation> metadata, String name,
          T Function(ElementAnnotation meta, DartObject instance) mapper) =>
      metadata.map((meta) {
        var value = meta.computeConstantValue()!;
        if (value.type?.element?.library?.name == 'controller.meta' &&
            value.type?.element?.name == name) {
          return mapper(meta, value);
        }
        return null;
      }).firstWhere((element) => element != null, orElse: () => null);

  HttpRequest? _extractHttpRequest(ElementAnnotation element) {
    var value = element.computeConstantValue()!;
    if (value.type?.element?.library?.name == 'controller.meta' &&
        httpRequestMeta.contains(value.type?.element?.name)) {
      return _constructHttpRequest(value);
    }
    return null;
  }

  HttpRequest _constructHttpRequest(DartObject value) {
    if (value.type?.element?.name == 'HttpRequest') {
      return HttpRequest(value.getField('method')!.toStringValue()!,
          value.getField('path')!.toStringValue()!);
    } else {
      return _constructHttpRequest(value.getField('(super)')!);
    }
  }

  String _buildMatcher(MethodElement element, HttpRequest httpRequest) {
    return '''
    .when2(
      matchRequest('${httpRequest.method}', '${httpRequest.path}'),
      _${element.name},
    )
    ''';
  }

  String _buildRequestHandler(
      MethodElement element, HttpRequest httpRequest, SecurityInfo? security) {
    var bindArguments = _bindArguments(element);
    var decodeBody = _buildDecodeBody(element);
    var checkSecurity = _buildSecurityCheck(security);
    var validate = _buildValidation(element);
    return '''
    Future<Response> _${element.name}(Map<String, String> headers, dynamic body) async {
      $checkSecurity
      $decodeBody
      $bindArguments
      $validate
      return toResponse(await controller.${element.name}(${_parameters(element)}));
    }
    ''';
  }

  String _buildDecodeBody(MethodElement element) {
    var decodeBody = '';
    var bodyParameters = element.parameters.where(_isBody);
    if (bodyParameters.isNotEmpty) {
      var bodyParameter = bodyParameters.first;
      var required = hasMeta(bodyParameter, Required);
      var bodyParameterName = bodyParameter.type.element!.name;
      decodeBody = 'var decodedBody = decodeJson(body);'
          'var jsonErrors = collectErrors(['
          '  await validJson.validate(\'body\', body),'
          '${required ? '  await required.validate(\'body\', decodedBody),' : ''}'
          '  await $bodyParameterName\$Validator.instance.validateDocument(decodedBody),'
          ']);'
          'if (jsonErrors.isNotEmpty) {'
          '  return badRequest(jsonErrors);'
          '}';
    }
    return decodeBody;
  }

  String _parameters(MethodElement element) =>
      element.parameters.map((parameter) {
        var converted = _convertIfNecessary(
          parameter.type,
          parameter.name,
          parameter.defaultValueCode,
        );
        return parameter.hasDefaultValue ||
                parameter.type.nullabilitySuffix != NullabilitySuffix.none
            ? '$converted,'
            : '$converted!,';
      }).join();

  String _buildSecurityCheck(SecurityInfo? secured) {
    return secured != null
        ? '''
      if (!(await security.verify(headers, ${secured.metadata.toSource().substring(1)}))) {
        return Response(${secured.statusCode});
      }
      '''
        : '';
  }

  String _bindArguments(MethodElement element) =>
      element.parameters.map(_bindArgument).join();

  String _bindArgument(ParameterElement element) => _isBody(element)
      ? 'var ${element.name} = ${_applyDefault('decodedBody', '${element.type.element?.name}.fromJson(decodedBody)', 'null', element.type.nullabilitySuffix)};\n'
      : 'var ${element.name} = ${"headers['${element.name}']"};\n';

  String _convertIfNecessary(
    DartType type,
    String argument,
    String? defaultValue,
  ) =>
      _applyDefault(argument, 'convert<${type.element?.name}>($argument)',
          defaultValue, type.nullabilitySuffix);

  String _applyDefault(
    String argument,
    String expression,
    String? defaultValue,
    NullabilitySuffix nullabilitySuffix,
  ) =>
      (argument == expression && defaultValue == null) ||
              nullabilitySuffix == NullabilitySuffix.none
          ? expression
          : '$argument != null ? $expression : $defaultValue';

  bool _isBody(ParameterElement element) => hasMeta(element, Body);

  String _buildValidation(MethodElement element) {
    var validators = element.parameters
        .expand((parameter) {
          var validators = <String?>{};
          _addEntityValidator(parameter, validators);
          _addCoreValidators(parameter, validators);
          _addCustomValidators(parameter, validators);
          return validators;
        })
        .where((validator) => validator != null)
        .toList();
    return validators.isNotEmpty
        ? '''
    var errors = collectErrors([${validators.join()}]);
    if (errors.isNotEmpty) {
      return badRequest(errors);
    }
    '''
        : '';
  }

  void _addCustomValidators(
      ParameterElement parameter, Set<String?> validators) {
    validators.addAll(parameter.metadata
        .where((metadata) =>
            isType(metadata.computeConstantValue()!.type!, Validator))
        .map((metadata) => 'await ${metadata.toSource().substring(1)}'
            ".validate('${parameter.name}', ${parameter.name}),"));
  }

  void _addCoreValidators(ParameterElement parameter, Set<String?> validators) {
    return coreValidators.forEach((check, validator) {
      if (check(parameter.type)) {
        validators.add(
            "await $validator.validate('${parameter.name}', ${parameter.name}),");
      }
    });
  }

  void _addEntityValidator(
      ParameterElement parameter, Set<String?> validators) {
    if (_isValidatable(parameter.type.element)) {
      validators.add(
          'await ${parameter.type.element!.name}\$Validator.instance.validateEntity(${parameter.name}),');
    }
  }

  bool _isValidatable(Element? element) =>
      element != null && hasMeta(element, Validatable);
}

extension NotNull<T> on Iterable<T?> {
  Iterable<T> whereNotNull() => where((e) => e != null).map((e) => e!);
}

class SecurityInfo {
  final int statusCode;
  final ElementAnnotation metadata;

  SecurityInfo(this.statusCode, this.metadata);
}
